#define I18N(x) _PSI.GetMsg(_PSI.ModuleNumber, static_cast<int>(x))

enum i18n_index {
ps__cancel,
ps__no,
ps__retry,
ps__yes,
ps_cancel,
ps_column,
ps_err_block_lines,
ps_err_block_type,
ps_err_block_whole1,
ps_err_block_whole2f,
ps_err_block_whole2l,
ps_err_block_whole3,
ps_err_block_whole4,
ps_helptopic,
ps_ok,
ps_reverse,
ps_title,
};
