#include <farplug-wide.h>
#include <assert.h>
#include <vector>
#include <string>
#include <map>

#include "i18nindex.h"

using namespace std;

extern PluginStartupInfo    _PSI;
extern FarStandardFunctions _FSF;
