#pragma once
enum {
M_TITLE,

M_OK,
M_CANCEL,
M_EXTRA_COMMANDS,
M_EXTRA_COMMANDS_TITLE,

M_INPUT_CMDNAME_TITLE,
M_INPUT_CMDNAME_PROMPT,

M_INPUT_CMDLINE_TITLE,
M_INPUT_CMDLINE_PROMPT,

M_HINT_NAVIGATE,
M_HINT_PAN,
M_HINT_SELECTION,
M_HINT_OTHER,

M_TEXT_USEORIENTATION,
M_TEXT_OPENBYENTER,
M_TEXT_OPENBYCTRLPGDN,
M_TEXT_OPENINQVIEW,
M_TEXT_OPENINFVIEW,

M_TEXT_IMAGEMASKS,
M_TEXT_VIDEOMASKS,

};
