#define I18N(x) _PSI.GetMsg(_PSI.ModuleNumber, static_cast<int>(x))

enum i18n_index {
ps__cancel,
ps__no,
ps__retry,
ps__yes,
ps_cancel,
ps_cfg_add_em,
ps_cfg_add_pm,
ps_cfg_add_vm,
ps_cfg_clr_active,
ps_cfg_clr_offset,
ps_cfg_clr_title,
ps_cfg_clr_updated,
ps_cfg_move_ib,
ps_cfg_prefix,
ps_cfg_save_pos,
ps_cfg_show_dd,
ps_cfg_std_csize,
ps_err_open_file,
ps_err_read_file,
ps_err_save_file,
ps_find_backward,
ps_find_empty,
ps_find_not_found,
ps_find_title,
ps_fn_codepage,
ps_fn_exit,
ps_fn_find,
ps_fn_goto,
ps_fn_help,
ps_fn_mode_ro,
ps_fn_mode_rw,
ps_fn_next,
ps_fn_prev,
ps_fn_save,
ps_fn_saveas,
ps_fn_screen,
ps_fn_setup,
ps_goto_offset,
ps_goto_out_ofr,
ps_goto_spec,
ps_goto_title,
ps_goto_topic,
ps_helptopic,
ps_ok,
ps_sav_alrexist,
ps_sav_cancelq,
ps_sav_file,
ps_sav_modifq,
ps_sav_overwrq,
ps_sav_readonly,
ps_sav_saveas,
ps_sav_title,
ps_swmod_quest,
ps_swmod_warn,
ps_title,
};
